/*!
 * Copyright (c) 2015-2019 Cisco Systems, Inc. See LICENSE file.
 */
/* globals navigator */
import bowser from 'bowser';

import LoggerProxy from '../common/logs/logger-proxy';
import {
  AUDIO_INPUT,
  VIDEO_INPUT,
  PEER_CONNECTION_STATE
} from '../constants';
import PeerConnectionManager from '../peer-connection-manager';
import ReconnectionError from '../common/errors/reconnection';
import MediaError from '../common/errors/media';
import config from '../config';


/**
 * MediaDirection
 * @typedef {Object} MediaDirection
 * @property {boolean} sendAudio
 * @property {boolean} receiveAudio
 * @property {boolean} sendVideo
 * @property {boolean} receiveVideo
 * @property {boolean} sendShare
 * @property {boolean} receiveShare
 */

/**
 * SendOptions
 * @typedef {Object} SendOptions
 * @property sendAudio
 * @property sendVideo
 * @property sendShare
 * @property {Object} sharePreferences
 */
/**
 *
 * @public
 * @export
 * Mimic browser APIs as "the ultimate browser".
 * Handles the quirks of each browser.
 * Extends and enhances adapter.js, i.e., the "media" file from the web client.
 */
const Media = {};

/**
 * @param {boolean} enabled
 * @param {MediaStreamTrack} track
 * @returns {Boolean}
 * @public
 */
Media.setLocalTrack = (enabled, track) => {
  if (track) {
    track.enabled = !enabled;

    return true;
  }

  return false;
};

/**
 * @param {RTCPeerConnection} peerConnection
 * @param {Object} meetingProperties
 * @param {string} meetingProperties.meetingId
 * @param {string} meetingProperties.remoteQualityLevel LOW|MEDIUM|HIGH
 * @returns {Promise}
 */
Media.reconnectMedia = (peerConnection, {meetingId, remoteQualityLevel}) => {
  if (peerConnection.connectionState === PEER_CONNECTION_STATE.CLOSED ||
    peerConnection.connectionState === PEER_CONNECTION_STATE.FAILED) {
    return Promise.reject(new ReconnectionError('Reinitiate peerconnection'));
  }

  return PeerConnectionManager.createOffer(peerConnection, {meetingId, remoteQualityLevel});
};

/**
 * format the media array for send
 * @param {String} mediaId
 * @param {Boolean} audioMuted
 * @param {Boolean} videoMuted
 * @returns {Array} medias
 */
Media.generateLocalMedias = (mediaId, audioMuted, videoMuted) => {
  if (mediaId) {
    return [
      {
        localSdp: JSON.stringify({
          audioMuted,
          videoMuted
        }),
        mediaId
      }
    ];
  }

  return [];
};

/**
 * make a browser call to get the media
 * @param {SendOptions} options
 * @returns {Promise}
 */
Media.getLocalMedia = (options) => {
  const {
    sendAudio, sendVideo, sendShare, sharePreferences
  } = options;

  if (sendAudio || sendVideo) {
    return Media.getMedia(sendAudio, sendVideo);
  }

  if (sendShare) {
    return Media.getDisplayMedia(
      {
        sendAudio: false,
        sendShare: true,
        sharePreferences
      }
    );
  }

  return Promise.resolve(undefined);
};
/**
 * Returns the direction and tracks
 * @param {string} trackType type of track (audio/video)
 * @param {object}  track  tracks passed
 * @param {boolean} receiveTracks do you want to receive tracks from the remote side
 * @returns {Object} returns direction tracks to be added in transceiver
 */
Media.checkTracks = (trackType, track, receiveTracks) => {
  const getDirection = (sendTracks, receiveTracks) => {
    if (sendTracks && receiveTracks) {
      return 'sendrecv';
    } if (sendTracks && !receiveTracks) {
      return 'sendonly';
    } if (!sendTracks && receiveTracks) {
      return 'recvonly';
    } if (sendTracks && receiveTracks) {
      return 'inactive';
    }

    return 'inactive';
  };

  if (track) {
    return {track, direction: getDirection(!!track, receiveTracks)};
  }

  return {track: trackType, direction: getDirection(!!track, receiveTracks)};
};
/**
 * creates peerconnection and attaches streams
 * @param {MediaDirection} mediaProperties
 * @param {Object} meetingProperties
 * @param {string} meetingProperties.meetingId
 * @param {string} meetingProperties.remoteQualityLevel LOW|MEDIUM|HIGH
 * @returns {Array} [peerConnection, ]
 */
Media.attachMedia = (mediaProperties, {meetingId, remoteQualityLevel}) => {
  const {
    mediaDirection, audioTrack, videoTrack, shareTrack, peerConnection
  } = mediaProperties;

  let result = null;

  // Add Transceiver for audio
  result = Media.checkTracks('audio', mediaDirection.sendAudio && audioTrack, mediaDirection.receiveAudio);
  peerConnection.audioTransceiver = peerConnection.addTransceiver(result.track, {direction: result.direction});

  // Add Transceiver for video
  result = Media.checkTracks('video', mediaDirection.sendVideo && videoTrack, mediaDirection.receiveVideo);
  peerConnection.videoTransceiver = peerConnection.addTransceiver(result.track, {direction: result.direction});

  // Add Transceiver for share
  result = Media.checkTracks('video', mediaDirection.sendShare && shareTrack, mediaDirection.receiveShare);
  peerConnection.shareTransceiver = peerConnection.addTransceiver(result.track, {direction: result.direction});

  peerConnection.onnegotiationneeded = (event) => {
    LoggerProxy.logger.info(`Media->attachMedia/onnegotiationneeded#PeerConnection: ${event}`);
  };

  return PeerConnectionManager.createOffer(peerConnection, {meetingId, remoteQualityLevel});
};

/**
 * updates all the media streams and creates a new media offer
 * @param {MediaDirection} mediaProperties
 * @param {Object} meetingProperties
 * @param {string} meetingProperties.meetingId
 * @param {string} meetingProperties.remoteQualityLevel LOW|MEDIUM|HIGH
 * @returns {Promise}
 */
Media.updateMedia = (mediaProperties, {meetingId, remoteQualityLevel}) => {
  const {
    mediaDirection, audioTrack, videoTrack, shareTrack, peerConnection
  } = mediaProperties;

  // update audio transceiver
  Media.setTrackOnTransceiver(peerConnection.audioTransceiver, {
    type: 'audio',
    track: audioTrack,
    sendTrack: mediaDirection.sendAudio && audioTrack,
    receiveTrack: mediaDirection.receiveAudio
  });

  // update video transceiver
  Media.setTrackOnTransceiver(peerConnection.videoTransceiver, {
    type: 'video',
    track: videoTrack,
    sendTrack: mediaDirection.sendVideo && videoTrack,
    receiveTrack: mediaDirection.receiveVideo
  });

  // update content transceiver
  Media.setTrackOnTransceiver(peerConnection.shareTransceiver, {
    type: 'video',
    track: shareTrack,
    sendTrack: mediaDirection.sendShare && shareTrack,
    receiveTrack: mediaDirection.receiveShare
  });
  peerConnection.onnegotiationneeded = (event) => {
    LoggerProxy.logger.info(`Media->attachMedia/onnegotiationneeded#PeerConnection: ${event}`);
  };

  return PeerConnectionManager.createOffer(peerConnection, {meetingId, remoteQualityLevel});
};

/**
 * @param {RTCRtpTransceiver} transceiver
 * @param {Object} options
 * @param {MediaStreamTrack} options.track
 * @returns {undefined}
 */
Media.setTrackOnTransceiver = (transceiver, options) => {
  const {
    type, track, sendTrack, receiveTrack
  } = options;

  try {
    const result = Media.checkTracks(type, sendTrack && track, receiveTrack);

    transceiver.direction = result.direction;
    if (options.track) {
      transceiver.sender.replaceTrack(track);
    }
  }
  catch (e) {
    throw e;
  }
};

/**
 * creates a new offer
 * @param {Object} meetingProperties
 * @param {string} meetingProperties.meetingId
 * @param {string} meetingProperties.remoteQualityLevel LOW|MEDIUM|HIGH
 * @param {RTCPeerConnection} peerConnection
 * @param {RTCRtpTransceiver} transceiver
 * @param {Object} options see #Media.setTrackOnTransceiver
 * @returns {Promise}
 */
Media.updateTransceiver = ({meetingId, remoteQualityLevel}, peerConnection, transceiver, options) => {
  Media.setTrackOnTransceiver(transceiver, options);

  return PeerConnectionManager.createOffer(peerConnection, {meetingId, remoteQualityLevel});
};

/**
 * generates share streams
 * @param {object} options parameter
 * @param {Object} options.sendAudio sendAudio: {Boolean} sendAudio constraints
 * @param {Object} options.sendShare sendShare: {Boolean} sendShare constraints
 * @param {Object} options.sharePreferences sharePreferences: {Object} Share constraints and share constraints
 * @returns {Object} {streams}
 */
Media.getDisplayMedia = (options) => {
  const shareConstraints = options.sharePreferences && options.sharePreferences.shareConstraints || {
    cursor: 'always',
    frameRate: options.sharePreferences && options.sharePreferences.highFrameRate ? config.meetings.videoShareFrameRate : config.meetings.screenFrameRate,
    aspectRatio: config.meetings.aspectRatio,
    width: options.sharePreferences && options.sharePreferences.highFrameRate ? config.meetings.resolution.idealWidth : config.meetings.screenResolution.idealWidth,
    height: options.sharePreferences && options.sharePreferences.highFrameRate ? config.meetings.resolution.idealHeight : config.meetings.screenResolution.idealHeight
  };
  // chrome and webkit based browsers (edge, safari) automatically adjust everything
  // and we have noticed higher quality with those browser types
  // firefox specifically has some issues with resolution and frame rate decision making
  // so we are making it optional and configurable (with defaults) for firefox
  // to have higher quality, and for developers to control the values
  // eventually we may have to add the same functionality to chrome, OR conversely, get to with firefox


  if (bowser.name.toLowerCase() === 'firefox') {
    const mediaConfig = {
      audio: options.sendAudio,
      video: options.sendShare
    };


    return navigator.mediaDevices.getDisplayMedia({audio: options.sendAudio, video: mediaConfig})
      .then((stream) => {
        if (options.sendShare && stream.getVideoTracks().length > 0) {
          // Firefox has a bug with the spec where changing in the height and width only happens
          // after we get the inital tracks
          // https://bugzilla.mozilla.org/show_bug.cgi?id=1321221
          stream.getVideoTracks()[0].applyConstraints(shareConstraints);
        }


        return stream;
      });
  }

  return navigator.mediaDevices.getDisplayMedia({audio: options.sendAudio, video: options.sendShare ? shareConstraints : false});
};

/**
 * generates audio and video using constraints (often called after getSupportedDevices)
 * @param {Object} audio audio: {Boolean} gum constraints
 * @param {Object} video video: {Boolean} gum constraints
 * @returns {Object} {streams}
 */
Media.getMedia = (audio, video) => {
  const defaultWidth = {ideal: config.meetings.resolution.idealWidth, max: config.meetings.resolution.maxWidth};
  const defaultHeight = {ideal: config.meetings.resolution.idealHeight, max: config.meetings.resolution.maxHeight};
  const mediaConfig = {
    audio,
    video: video ? {
      deviceId: video.deviceId ? video.deviceId : undefined,
      width: video.width ? video.width : defaultWidth,
      height: video.height ? video.height : defaultHeight
    } : false,
    fake: process.env.NODE_ENV === 'test' // Special case to get fake media for Firefox browser for testing
  };

  return new Promise((resolve, reject) => {
    navigator.mediaDevices
      .getUserMedia(mediaConfig)
      .then((stream) => {
        resolve(stream);
      })
      .catch((err) => {
        reject(
          new MediaError('Local Get User Media failed', {
            name: err.name,
            localMediaFailed: true,
            cameraFailed: err === mediaConfig.permissionDenied
          })
        );
      });
  });
};

/**
 * Checks if the machine has at least one audio or video device (Dont use this for screen share)
 * @param {object} [options]
 * {
 *    sendAudio: true/false,
 *    sendVideo: true/false
 * }
 * @returns {Object} {
 *    sendAudio: true/false,
 *    sendVideo: true/false
 *}
 */
Media.getSupportedDevice = ({sendAudio, sendVideo}) =>
  Promise.resolve().then(() => {
    if (!navigator.mediaDevices || navigator.mediaDevices.enumerateDevices === undefined) {
      return {
        sendAudio: true,
        sendVideo: true
      };
    }

    return navigator.mediaDevices.enumerateDevices().then((devices) => {
      const supported = {
        audio: devices.filter((device) => device.kind === AUDIO_INPUT).length > 0,
        video: devices.filter((device) => device.kind === VIDEO_INPUT).length > 0
      };

      return {
        sendAudio: (supported.audio && sendAudio) || true,
        sendVideo: (supported.video && sendVideo) || true
      };
    });
  });

/**
 * proxy to browser navigator.mediaDevices.enumerateDevices()
 * @returns {Promise}
 */
Media.getDevices = () => {
  if (navigator && navigator.mediaDevices && navigator.mediaDevices.enumerateDevices) {
    return navigator.mediaDevices.enumerateDevices();
  }

  return Promise.reject(new MediaError('enumerateDevices not supported.'));
};

/**
 *
 * Toggle a specific stream
 * noop as of now, does nothing
 * @returns {null}
 */
Media.toggleStream = () => {};

/**
 * Stop input stream
 * @param {MediaTrack} track A media stream
 * @returns {null}
 */
Media.stopTracks = (track) => {
  if (!track) {
    return Promise.resolve();
  }

  return Promise.resolve().then(() => {
    if (track && track.stop) {
      try {
        track.stop();
      }
      catch (e) {
        LoggerProxy.logger.error(`media/index->stopTracks#Unable to stop the track with state ${track.readyState}, error: ${e}`);
      }
    }
  });
};

/**
 *
 * Stop input stream
 * @param {Stream} stream A media stream
 * @returns {null}
 */
Media.stopStream = (stream) => {
  if (!stream) {
    return Promise.resolve();
  }

  /*
   * To release local media
   * 1) Chrome requires all tracks to be stopped (stream.stop got deprecated)
   * 2) Firefox requires the stream to be stopped
   */
  return Promise.resolve().then(() => {
    if (stream.getTracks) {
      stream.getTracks().forEach((track) => {
        track.stop();
      });
    }
    else if (stream.stop) {
      stream.stop();
    }
  });
};

/**
 * generates streams for audio video and share
 * @param {object} mediaSetting parameter
 * @param {Object} mediaSetting.sendAudio sendAudio: {Boolean} sendAudio constraints
 * @param {Object} mediaSetting.sendVideo sendVideo: {Boolean} sendVideo constraints
 * @param {Object} mediaSetting.sendShare sendShare: {Boolean} sendShare constraints
 * @param {Object} audioVideo parameter
 * @param {Object} audioVideo.audio {deviceId: {String}}
 * @param {Object} audioVideo.video {deviceId: {String}}
 * @param {Object} sharePreferences parameter
 * @param {Object} sharePreferences.shareConstraints parameter
 * @param {Boolean} sharePreferences.highFrameRate parameter
 * @returns {Array} [localStream, shareStream]
 */
Media.getUserMedia = (mediaSetting, audioVideo, sharePreferences) => Media.getLocalMedia({
  sendAudio: mediaSetting.sendAudio ? audioVideo.audio || mediaSetting.sendAudio : false,
  sendVideo: mediaSetting.sendVideo ? audioVideo.video || mediaSetting.sendVideo : false
}).then((localStream) => Media.getLocalMedia({
  sendShare: mediaSetting.sendShare,
  sharePreferences
}).then((shareStream) => [localStream, shareStream]));

export default Media;
